use base64::prelude::*;
use eyre::{Result, WrapErr};
use indicatif::{ProgressBar, ProgressStyle};
use std::{
    ffi::OsString,
    fmt::{Debug, Display},
    io::{BufRead, BufReader, Read, Write},
    os::unix::net::UnixStream,
    path::{Path, PathBuf},
    process::{Command, Stdio},
    time::{Duration, SystemTime, UNIX_EPOCH},
};

const CHUNK_SIZE: usize = 1024 * 16;

fn spinner(msg: impl Display) -> ProgressBar {
    cfg_if::cfg_if! {
        if #[cfg(feature = "unicode")] {
            let style = ProgressStyle::default_spinner();
            let delay = Duration::from_millis(100);
        } else {
            let style = ProgressStyle::default_spinner().tick_chars(r#"\|/-!"#);
            let delay = Duration::from_millis(200);
        }
    };
    let bar = ProgressBar::new_spinner().with_style(style);
    bar.enable_steady_tick(delay);
    bar.set_message(msg.to_string());
    bar
}

fn bar(count: u64, msg: impl Display) -> ProgressBar {
    let template = "{elapsed_precise} [{wide_bar}] {percent}% {msg}";
    cfg_if::cfg_if! {
        if #[cfg(feature = "unicode")] {
            let style = ProgressStyle::with_template(template).unwrap();
        } else {
            let style = ProgressStyle::with_template(template).unwrap().progress_chars("=> ");
        }
    }
    let bar = ProgressBar::new(count).with_style(style);
    bar.set_message(msg.to_string());
    bar
}

fn get_pid(path: impl AsRef<Path>) -> Result<u32> {
    let path = path.as_ref();
    let pid_str = std::fs::read_to_string(path).context("could not read PID")?;
    pid_str.parse().context("could not parse PID")
}

fn remove_if_exists(path: impl AsRef<Path>) -> Result<()> {
    if let Err(e) = std::fs::remove_file(path) {
        if e.kind() != std::io::ErrorKind::NotFound {
            return Err(e).context("could not remove stale file");
        }
    }
    Ok(())
}

fn to_lowercase_hexlike(s: impl AsRef<str>) -> String {
    let mut s = s.as_ref().trim();
    if s.starts_with("0x") {
        s = &s[2..];
    }
    s.to_ascii_lowercase()
}

#[allow(clippy::struct_field_names)]
#[derive(Clone, Debug)]
struct Device {
    vendor_id: u16,
    device_id: u16,
    bus_id: String,
}

fn find_pci_device_by_class(class: u16) -> Result<Vec<Device>> {
    let devices = std::fs::read_dir("/sys/bus/pci/devices")?;
    let mut valid_devices = vec![];
    for entry in devices {
        let entry = entry?;
        let path = entry.path();
        let bus_address = entry.file_name();
        let class_string = std::fs::read_to_string(path.join("class"))?;
        if to_lowercase_hexlike(class_string)[..4] == hex::encode(class.to_be_bytes()) {
            let device_id = u16::from_be_bytes(
                hex::decode(to_lowercase_hexlike(std::fs::read_to_string(
                    path.join("device"),
                )?))?
                .try_into()
                .map_err(|e| eyre::eyre!("could not convert to u16: {e:?}"))?,
            );
            let vendor_id = u16::from_be_bytes(
                hex::decode(to_lowercase_hexlike(std::fs::read_to_string(
                    path.join("vendor"),
                )?))?
                .try_into()
                .map_err(|e| eyre::eyre!("could not convert to u16: {e:?}"))?,
            );

            let bus_id = bus_address
                .into_string()
                .map_err(|bad| eyre::eyre!("non-utf8 bus address: {bad:?}"))?
                .split_once(':')
                .ok_or(eyre::eyre!("bad path ID"))?
                .1
                .to_string();
            valid_devices.push(Device {
                vendor_id,
                device_id,
                bus_id,
            });
        }
    }

    Ok(valid_devices)
}

// NOTE: Do not implement `clone`, as there is side-effect state involved.

/// A control handle for a virtual machine.
#[derive(Debug)]
pub struct VirtualMachine {
    pid: u32,
    // qemu guest agent (proxied to guest)
    guest_writer: UnixStream,
    guest_reader: BufReader<UnixStream>,
    // qemu machine protocol (host)
    host_writer: UnixStream,
    host_reader: BufReader<UnixStream>,
    args: SpawnArguments,
}

/// The configuration to use when starting a VM.
#[derive(Clone, Debug)]
pub struct SpawnArguments {
    /// The amount of memory to allocate to a VM.
    pub memory: String,

    /// The image file to use when booting the machine.
    ///
    /// By default, this is "/guest.img".
    pub guest_image: PathBuf,

    /// The path for the guest agent socket.
    pub guest_agent_socket_path: PathBuf,

    /// The path for the QMP socket.
    pub qmp_socket_path: PathBuf,

    /// The path for the lockfile.
    pub lockfile_path: PathBuf,
}

impl Default for SpawnArguments {
    fn default() -> Self {
        Self {
            memory: String::from("1G"),
            guest_image: PathBuf::from("/guest.img"),
            guest_agent_socket_path: PathBuf::from("/var/run/netvm_qga.sock"),
            qmp_socket_path: PathBuf::from("/var/run/netvm_qmp.sock"),
            lockfile_path: PathBuf::from("/var/run/netvm.lock"),
        }
    }
}

impl VirtualMachine {
    /// Start a virutal machine with the given parameters.
    pub fn start(args: SpawnArguments) -> eyre::Result<Self> {
        let eth_devices = find_pci_device_by_class(0x0200)?;

        // Ensure VM isn't already started
        if std::fs::exists(&args.lockfile_path)? {
            // Check if VM is running
            use nix::unistd::{getpgid, Pid};
            let pid = i32::try_from(get_pid(&args.lockfile_path)?)
                .context(eyre::eyre!("bad PID: pid < i32::MAX << 1"))?;
            if getpgid(Some(Pid::from_raw(pid))).is_ok() {
                // process exists, exit
                return Err(eyre::eyre!(
                    "VM with this configuration exists as PID {pid}"
                ));
            } else {
                // Remove old state
                eprintln!("Found stale lockfile (PID terminated), removing all state");
                remove_if_exists(&args.guest_agent_socket_path)?;
                remove_if_exists(&args.qmp_socket_path)?;
                remove_if_exists(&args.lockfile_path)?;
            }
        }

        let mut chardev_opts = OsString::from("socket,path=");
        chardev_opts.push(args.guest_agent_socket_path.as_os_str());
        chardev_opts.push(",server=on,wait=off,id=qga0");

        let mut qmp_opts = OsString::from("unix:");
        qmp_opts.push(args.qmp_socket_path.as_os_str());
        qmp_opts.push(",server,nowait");

        // TODO: https://git.distrust.co/public/airgap/src/commit/73ab8eae21898d80625062d6037e84dff61fabf8/src/host/rootfs/usr/local/bin/netvm#L90-L100
        // Add these options if necessary.

        let bar = spinner("Loading VM");

        if !eth_devices.is_empty() {
            std::fs::write(
                "/sys/module/vfio_iommu_type1/parameters/allow_unsafe_interrupts",
                "Y",
            )?;
        }

        let mut net_args = vec![];
        for dev in eth_devices {
            bar.println(format!("Attaching Ethernet device: {dev:?}"));
            let bus_id = &dev.bus_id;
            std::fs::write(
                "/sys/bus/pci/drivers/vfio-pci/new_id",
                format!(
                    "{} {}",
                    hex::encode(dev.vendor_id.to_be_bytes()),
                    hex::encode(dev.device_id.to_be_bytes())
                ),
            )?;
            net_args.push("-device".to_string());
            net_args.push(format!("vfio-pci,host={bus_id}"));
        }

        let mut child = Command::new("qemu-system-x86_64")
            .stdin(Stdio::null())
            .stdout(Stdio::null())
            .stderr(Stdio::null())
            .args(["-m", &args.memory])
            .args(["-machine", "q35"])
            .arg("-nographic")
            .args(["-serial", "none"])
            .args(["-monitor", "none"])
            .args(["-net", "none"])
            .arg("-cdrom")
            .arg(&args.guest_image)
            .args(["-boot", "order=d"])
            .arg("-chardev")
            .arg(chardev_opts)
            .arg("-qmp")
            .arg(qmp_opts)
            .args(net_args)
            .args(["-device", "qemu-xhci,id=usb"])
            .args(["-device", "virtio-serial"])
            .args([
                "-device",
                "virtserialport,chardev=qga0,name=org.qemu.guest_agent.0",
            ])
            .spawn()
            .context("unable to spawn qemu vm")?;

        std::fs::write(&args.lockfile_path, child.id().to_string())
            .context("could not write PID to file")?;

        loop {
            // Check if the child has exited prematurely, and if so, exit
            let result = child
                .try_wait()
                .context("error checking child process status")?;
            if let Some(status) = result {
                return Err(eyre::eyre!("child exited with code {:?}", status.code()));
            }

            if std::fs::exists(&args.guest_agent_socket_path)?
                && std::fs::exists(&args.qmp_socket_path)?
            {
                break;
            }
            std::thread::sleep(Duration::from_millis(100));
        }

        bar.finish_and_clear();

        Self::load(args, Some(child.id()))
    }

    /// Load a virtual machine with the given parameters and optionally a custom PID.
    ///
    /// The custom PID option may be relevant if the virtual machine sockets were loaded but the
    /// PID of the virtual machine was not properly persisted.
    pub fn load(args: SpawnArguments, pid: Option<u32>) -> Result<Self> {
        let bar = spinner("Connecting to VM");
        let pid = if let Some(pid) = pid {
            pid
        } else {
            let pid_str = std::fs::read_to_string(&args.lockfile_path)
                .context("error reading PID from lockfile")?;
            pid_str.parse().context("could not parse PID")?
        };

        let guest = UnixStream::connect(&args.guest_agent_socket_path)
            .context("could not open socket to QVM guest agent")?;

        let host = UnixStream::connect(&args.qmp_socket_path)
            .context("could not open socket to qemu management socket")?;

        bar.println(format!(
            "Connected to VM with PID {} and socket {}",
            pid,
            &args.guest_agent_socket_path.display(),
        ));
        bar.finish_and_clear();

        let vm = Self::from_parts(pid, guest, host, args)?;

        Ok(vm)
    }

    fn from_parts(
        pid: u32,
        guest_socket: UnixStream,
        host_socket: UnixStream,
        args: SpawnArguments,
    ) -> Result<Self> {
        let guest_reader = BufReader::new(
            guest_socket
                .try_clone()
                .context("couldn't clone socket to make buffered reader")?,
        );
        let mut host_reader = BufReader::new(
            host_socket
                .try_clone()
                .context("couldn't clone socket to make buffered reader")?,
        );

        let mut server_hello = String::new();
        host_reader
            .read_line(&mut server_hello)
            .context("can't read line from socket (pre-load)")?;

        let mut vm = Self {
            pid,
            guest_writer: guest_socket,
            guest_reader,
            host_writer: host_socket,
            host_reader,
            args,
        };

        vm.flush()?;
        // NOTE: it is fine to use the system time here, modulo u32 for sending over
        // the wire. this is because this is not meant to be _secure_, it's just meant
        // to be _unique_, and we can assume this machine will not be running for 2^32
        // seconds, and _definitely_ won't be having a collision. i'm fine with it
        // crashing if those circumstances happen to be met.
        let time = SystemTime::now().duration_since(UNIX_EPOCH)?;

        let identifier = time.as_secs() % u64::from(u32::MAX);

        let ping_response = vm
            .execute_internal("guest-sync", serde_json::json!({"id": identifier}))
            .context("couldn't ping")?;

        if ping_response.as_u64().is_none_or(|id| id != identifier) {
            return Err(eyre::eyre!(
                "known id {identifier} != given identifier {ping_response:?}"
            ));
        }

        Ok(vm)
    }

    /// The PID of the virtual machine.
    pub fn pid(&self) -> u32 {
        self.pid
    }

    fn flush(&mut self) -> Result<()> {
        // flush steps:
        // * put the reader into nonblocking mode
        // * read all the data possible to chew what's left over
        // * put the reader back into blocking mode
        // * put the parser in a bad state to reset it
        // * read a line from the parser to reset the input

        let bar = spinner("Re-establishing connection...");
        self.guest_writer
            .set_nonblocking(true)
            .context("flush: can't set nonblocking")?;
        if let Err(e) = self.guest_reader.read_to_end(&mut vec![]) {
            if e.kind() != std::io::ErrorKind::WouldBlock {
                return Err(e).context("flush: can't read nonblocked data");
            }
        }
        self.guest_writer
            .set_nonblocking(false)
            .context("flush: can't set blocking")?;
        self.guest_writer
            .write_all(&[0x1b])
            .context("flush: can't send reset byte")?;
        self.guest_reader
            .read_line(&mut String::new())
            .context("flush: can't read error")?;
        bar.finish_and_clear();

        Ok(())
    }

    /// Push a single file from the local machine to the VM.
    pub fn push(
        &mut self,
        local_path: impl AsRef<Path>,
        remote_path: impl AsRef<Path>,
    ) -> Result<()> {
        let remote_path_as_str = remote_path
            .as_ref()
            .to_str()
            .ok_or(eyre::eyre!("non-utf8 remote path filename"))?;
        let metadata = std::fs::metadata(&local_path)?;
        let bar = bar(
            metadata.len(),
            format!(
                "cp {local_path} vm:{remote_path}",
                local_path = local_path.as_ref().display(),
                remote_path = remote_path_as_str,
            ),
        );
        let mut local_reader = std::fs::File::open(local_path)?;
        let remote_handle = self
            .execute_internal(
                "guest-file-open",
                serde_json::json!({
                    "path": remote_path_as_str,
                    "mode": "w",
                }),
            )
            .context("could not open file")?
            .as_u64()
            .ok_or(eyre::eyre!("response was non-u64"))?;

        // put this on the heap, we don't want smash the stack.
        let mut buf = vec![0u8; CHUNK_SIZE];
        while let Ok(size) = local_reader.read(&mut buf) {
            bar.inc(size as u64);
            if size == 0 {
                break;
            }
            let mut written = 0usize;
            loop {
                let response = self.execute_internal(
                    "guest-file-write",
                    serde_json::json!({
                        "handle": remote_handle,
                        "buf-b64": BASE64_STANDARD.encode(&buf[written..size]),
                    }),
                )?;
                let response_written = response
                    .get("count")
                    .ok_or(eyre::eyre!("not given 'count' of bytes written"))?
                    .as_u64()
                    .ok_or(eyre::eyre!("'count' not u64"))?;
                written +=
                    usize::try_from(response_written).expect("wrote more than u46::MAX bytes");
                if written == size {
                    break;
                }
            }
        }

        self.execute_internal(
            "guest-file-close",
            serde_json::json!({"handle": remote_handle}),
        )?;
        bar.finish_and_clear();

        Ok(())
    }

    /// Pull a single file from the VM to the local machine. This operation is destructive and will
    /// overwrite existing files.
    pub fn pull(
        &mut self,
        remote_path: impl AsRef<Path>,
        local_path: impl AsRef<Path>,
    ) -> Result<()> {
        let remote_path_as_str = remote_path
            .as_ref()
            .to_str()
            .ok_or(eyre::eyre!("non-utf8 remote path filename"))?;
        let spinner = spinner("finding remote file");
        let remote_handle = self
            .execute_internal(
                "guest-file-open",
                serde_json::json!({
                    "path": remote_path_as_str,
                    "mode": "r",
                }),
            )
            .context("could not open file")?
            .as_u64()
            .ok_or(eyre::eyre!("response was non-u64"))?;
        let response = self.execute_internal(
            "guest-file-seek",
            serde_json::json!({
                "handle": remote_handle,
                "offset": 0,
                "whence": "end",
            }),
        )?;
        let size = response
            .get("position")
            .ok_or(eyre::eyre!("not given 'position' of bytes"))?
            .as_u64()
            .ok_or(eyre::eyre!("'position' not u64"))?;

        self.execute_internal(
            "guest-file-close",
            serde_json::json!({"handle": remote_handle}),
        )?;
        spinner.finish_and_clear();

        let bar = bar(
            size,
            format!(
                "cp vm:{remote_path} {local_path}",
                remote_path = remote_path_as_str,
                local_path = local_path.as_ref().display()
            ),
        );
        let mut local_writer = std::fs::File::create(local_path)?;
        let remote_handle = self
            .execute_internal(
                "guest-file-open",
                serde_json::json!({
                    "path": remote_path_as_str,
                    "mode": "r",
                }),
            )
            .context("could not open file")?
            .as_u64()
            .ok_or(eyre::eyre!("response was non-u64"))?;
        let mut buf = vec![];
        loop {
            let chunk = self.execute_internal(
                "guest-file-read",
                serde_json::json!({
                    "handle": remote_handle,
                }),
            )?;
            let base64_data = chunk
                .get("buf-b64")
                .ok_or(eyre::eyre!("not givenu 'buf-b64'"))?
                .as_str()
                .ok_or(eyre::eyre!("'buf-b64' not str"))?;
            BASE64_STANDARD.decode_vec(base64_data, &mut buf)?;
            local_writer.write_all(&buf)?;
            bar.inc(buf.len() as u64);
            buf.clear();
            if chunk
                .get("eof")
                .is_some_and(|eof| eof.as_bool().is_some_and(|b| b))
            {
                break;
            }
        }

        self.execute_internal(
            "guest-file-close",
            serde_json::json!({"handle": remote_handle}),
        )?;
        bar.finish_and_clear();

        Ok(())
    }

    // TODO: make this return status, stdout, stderr
    // TODO: accept optional: env, input-data, disable capture-output
    /// Run a command on the virtual machine. Standard input is not sent to the process, and only
    /// standard output is received from the process.
    pub fn run_command(
        &mut self,
        command: &str,
        args: impl IntoIterator<Item = String>,
        stdin: Option<&[u8]>,
    ) -> Result<(Vec<u8>, u64)> {
        let args = args.into_iter().collect::<Vec<_>>();
        let input = stdin.map(|bytes| BASE64_STANDARD.encode(bytes));

        let payload = serde_json::json!({
            "path": command,
            "arg": args,
            "capture-output": true,
            "input-data": input,
        });

        let bar = spinner(format!("Running: {command:?} {args:?}"));
        let result = self.execute_internal("guest-exec", &payload)?;
        let pid = result
            .get("pid")
            .ok_or(eyre::eyre!("no PID: {result:?}"))?
            .as_u64()
            .ok_or(eyre::eyre!("pid is not u64"))?;

        let payload = serde_json::json!({"pid": pid});
        let mut status;
        loop {
            status = self.execute_internal("guest-exec-status", &payload)?;
            if status
                .get("exited")
                .is_some_and(|e| e.as_bool().is_some_and(|b| b))
            {
                break;
            }

            std::thread::sleep(Duration::from_secs(1));
        }
        bar.finish_and_clear();

        let out_data = status
            .get("out-data")
            .ok_or(eyre::eyre!("response had no out-data"))?
            .as_str()
            .ok_or(eyre::eyre!("response['out-data'] is not string"))?;
        let parsed_data = BASE64_STANDARD
            .decode(out_data)
            .context("response output was not base64")?;
        let exit_code = status
            .get("exitcode")
            .ok_or(eyre::eyre!("response had no exitcode"))?
            .as_number()
            .ok_or(eyre::eyre!("response['exitcode'] is not number"))?
            .as_u64()
            .ok_or(eyre::eyre!("response['exitcode'] is not integer"))?;

        Ok((parsed_data, exit_code))
    }

    fn execute_internal<S: serde::Serialize>(
        &mut self,
        command: &'static str,
        args: S,
    ) -> Result<serde_json::Value> {
        let message = serde_json::json!({
            "execute": command,
            "arguments": args,
        });

        serde_json::to_writer(&mut self.guest_writer, &message)
            .context("could not send message over socket")?;
        writeln!(&mut self.guest_writer).context("could not send newline over socket")?;

        let mut line = String::new();
        self.guest_reader
            .read_line(&mut line)
            .context("can't read line from socket")?;

        let response: serde_json::Value =
            serde_json::from_str(&line).context("response from qemu-guest-agent is not json")?;

        if let Some(response) = response.get("return") {
            Ok(response.clone())
        } else if let Some(error) = response.get("error") {
            Err(eyre::eyre!("error response from qemu: {error:?}"))
        } else {
            Err(eyre::eyre!("invalid response from qemu: {response:?}"))
        }
    }

    /// Execute an operation via QEMU Guest Agent. This modifies state inside the VM.
    pub fn execute<S: serde::Serialize + Debug>(
        &mut self,
        command: &'static str,
        args: S,
    ) -> Result<serde_json::Value> {
        let bar = spinner(format!("Executing: {command:?} with {args:?}"));
        let result = self.execute_internal(command, args);
        bar.finish_and_clear();
        result
    }

    /// Execute an operation via QEMU Machine Protocol. This modifies state on the host machine and
    /// the VM.
    pub fn execute_host<S: serde::Serialize + Debug>(
        &mut self,
        command: &'static str,
        args: S,
    ) -> Result<serde_json::Value> {
        let bar = spinner(format!("Executing: {command:?} with {args:?}"));

        let message = serde_json::json!({
            "execute": command,
            "arguments": args,
        });

        serde_json::to_writer(&mut self.host_writer, &message)
            .context("could not send message over socket")?;
        writeln!(&mut self.host_writer).context("could not send newline over socket")?;

        let mut line = String::new();
        self.host_reader
            .read_line(&mut line)
            .context("can't read line from socket")?;

        let response: serde_json::Value =
            serde_json::from_str(&line).context("response from qemu is not json")?;

        bar.finish_and_clear();

        if let Some(response) = response.get("return") {
            Ok(response.clone())
        } else if let Some(error) = response.get("error") {
            Err(eyre::eyre!("error response from qemu: {error:?}"))
        } else {
            Err(eyre::eyre!("invalid response from qemu: {response:?}"))
        }
    }

    // NOTE: u32 is returned from Process::id(), i32 is the Linux internal version
    // This should be safe; the kernel wouldn't give a value that, when converted
    // to a u32, can't be made back into an i32
    /// Terminate the VM and remove any stateful files.
    pub fn kill(self) -> Result<()> {
        use nix::{
            errno::Errno,
            sys::signal::{kill, SIGKILL},
            unistd::{getpgid, Pid},
        };
        let pid = Pid::from_raw(
            i32::try_from(self.pid).context(eyre::eyre!("bad PID: pid < i32::MAX << 1"))?,
        );
        if getpgid(Some(pid)).is_err() {
            eprintln!("Process not found");
            return Ok(());
        }

        let bar = spinner("Killing existing PID");
        loop {
            if let Err(e) = kill(pid, SIGKILL) {
                if e == Errno::ESRCH {
                    break;
                }
            }
            std::thread::sleep(Duration::from_millis(50));
        }
        /*
        // NOTE: `waitpid()` will not work for non-child processes
        // instead, continuously kill(pid, SIGKILL)
        let result = waitpid(pid, Some(WaitPidFlag::WEXITED));
        if let Err(e) = result {
            if e != Errno::ECHILD {
                return Err(eyre::eyre!("couldn't await child {pid} (os error {e})"));
            }
        }
        */
        remove_if_exists(&self.args.guest_agent_socket_path)?;
        remove_if_exists(&self.args.qmp_socket_path)?;
        remove_if_exists(&self.args.lockfile_path)?;
        bar.finish_and_clear();
        Ok(())
    }
}
